import os

ALLOWED_ORIGINS = os.environ.get("ALLOWED_ORIGINS", "*").split()
ISCC_STREAM = os.environ.get("ISCC_STREAM", "iscc")
NODE_IP = os.environ.get("NODE_IP", "127.0.0.1")
NODE_PORT = os.environ.get("NODE_PORT", "9718")
NODE_USER = os.environ.get("NODE_USER", "")
NODE_PWD = os.environ.get("NODE_PWD", "")



# -*- coding: utf-8 -*-
from datetime import datetime
from typing import Optional, List

from pydantic import BaseModel, Field, HttpUrl


def main():
    pass


if __name__ == "__main__":
    main()


class Metadata(BaseModel):
    """Metadata for Meta-ID creation."""

    title: str = Field(
        ..., description="The title of an intangible creation.", min_length=1
    )
    extra: Optional[str] = Field(
        None,
        description="An optional short statement that distinguishes "
        "this intangible creation from another one for the "
        "purpose of forced Meta-ID uniqueness.",
    )


class Text(BaseModel):
    text: str = Field(None, description="Extracted full plain text for Content-ID.")


class ISCC(BaseModel):
    """Full ISCC Code including Metadata."""

    iscc: str = Field(..., description="Full ISCC Code")
    title: str = Field(None, description="Title of intangible creation")
    title_trimmed: str = Field(None, description="Normalized and trimmed title")
    extra: str = Field(None, description="Optional extra metadata")
    extra_trimmed: str = Field(
        None, description="Normalized and trimmed extra metadata"
    )
    tophash: str = Field(..., description="Normalized Title")
    gmt: str = Field(..., description="Generic Media Type")
    bits: List[str] = Field(..., description="Per component bitstrings")


class IsccComponent(BaseModel):
    """A single ISCC Component as code, bits and headerless integer."""

    code: str = Field(..., description="Single ISCC component", max_length=13)
    bits: str = Field(..., description="Bitstring of component body", max_length=64)
    ident: int = Field(
        ..., description="Integer representation of component body", le=2 ** 64
    )


class MetaID(IsccComponent):
    """A Meta-ID ISCC Component including Metadata."""

    title: str = Field(..., description="Title of intangible creation")
    title_trimmed: str = Field(..., description="Normalized and trimmed title")
    extra: str = Field(None, description="Optional extra metadata")
    extra_trimmed: str = Field(
        None, description="Normalized and trimmed extra metadata"
    )


class ContentID(IsccComponent):
    """A Content-ID ISCC Component including Generic Media Type."""

    gmt: str = Field(
        "text", description="Generic Media Type of Content-ID", max_length=64
    )


class DataID(IsccComponent):
    """A Data-ID ISCC Component."""

    pass


class InstanceID(IsccComponent):
    """An Instance-ID ISCC Component including Tophash."""

    tophash: str = Field(
        None, description="Hex-encoded 256-bit Top Hash", max_length=64
    )


class StreamItem(BaseModel):
    txid: str
    vout: int
    keys: List[str]
    title: str
    extra: Optional[str]
    tophash: Optional[str]
    time: datetime
    content_url: Optional[HttpUrl]
    bits: List[str]



import os
import shutil
import uuid
from os.path import join, splitext
from typing import List

import mobi
import uvicorn
from fastapi import FastAPI, UploadFile, File, Form, HTTPException
import iscc
from tika import detector, parser
from iscc_cli.const import SUPPORTED_MIME_TYPES, GMT
import iscc_service
from iscc_service.config import ALLOWED_ORIGINS, ISCC_STREAM
from iscc_service.conn import get_client
from iscc_service.models import (
    Metadata,
    Text,
    ISCC,
    MetaID,
    ContentID,
    DataID,
    InstanceID,
    StreamItem,
)
from iscc_service.tools import (
    code_to_bits,
    code_to_int,
    stream_filter,
    add_placeholders,
)
from pydantic import HttpUrl
from iscc_cli.lib import iscc_from_url
from iscc_cli.utils import iscc_split, get_title, mime_to_gmt, iscc_verify
from iscc_cli import APP_DIR, audio_id, video_id
from starlette.middleware.cors import CORSMiddleware
from starlette.status import (
    HTTP_415_UNSUPPORTED_MEDIA_TYPE,
    HTTP_422_UNPROCESSABLE_ENTITY,
    HTTP_503_SERVICE_UNAVAILABLE,
    HTTP_400_BAD_REQUEST,
)


app = FastAPI(
    title="ISCC Web Service API",
    version=iscc_service.__version__,
    description="Microservice for creating ISCC Codes from Media Files.",
    docs_url="/",
)

app.add_middleware(
    CORSMiddleware,
    allow_origins=ALLOWED_ORIGINS,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.post(
    "/generate/from_file",
    response_model=ISCC,
    response_model_exclude_unset=True,
    tags=["generate"],
    summary="Generate ISCC from File",
)
def from_file(
    file: UploadFile = File(...), title: str = Form(""), extra: str = Form("")
):
    """Generate Full ISCC Code from Media File with optional explicit metadata."""

    media_type = detector.from_buffer(file.file)
    if media_type not in SUPPORTED_MIME_TYPES:
        raise HTTPException(
            HTTP_415_UNSUPPORTED_MEDIA_TYPE,
            "Unsupported media type '{}'. Please request support at "
            "https://github.com/iscc/iscc-service/issues.".format(media_type),
        )

    if media_type == "application/x-mobipocket-ebook":
        file.file.seek(0)
        tempdir, filepath = mobi.extract(file.file)
        tika_result = parser.from_file(filepath)
        shutil.rmtree(tempdir)
    else:
        file.file.seek(0)
        tika_result = parser.from_buffer(file.file)

    if not title:
        title = get_title(tika_result, guess=True)

    mid, norm_title, norm_extra = iscc.meta_id(title, extra)
    gmt = mime_to_gmt(media_type)
    if gmt == GMT.IMAGE:
        file.file.seek(0)
        cid = iscc.content_id_image(file.file)
    elif gmt == GMT.TEXT:
        text = tika_result["content"]
        if not text:
            raise HTTPException(HTTP_422_UNPROCESSABLE_ENTITY, "Could not extract text")
        cid = iscc.content_id_text(tika_result["content"])
    elif gmt == GMT.AUDIO:
        file.file.seek(0)
        features = audio_id.get_chroma_vector(file.file)
        cid = audio_id.content_id_audio(features)
    elif gmt == GMT.VIDEO:
        file.file.seek(0)
        _, ext = splitext(file.filename)
        fn = "{}{}".format(uuid.uuid4(), ext)
        tmp_path = join(APP_DIR, fn)
        with open(tmp_path, "wb") as buffer:
            shutil.copyfileobj(file.file, buffer)
        features = video_id.get_frame_vectors(tmp_path)
        cid = video_id.content_id_video(features)
        os.remove(tmp_path)

    file.file.seek(0)
    did = iscc.data_id(file.file)
    file.file.seek(0)
    iid, tophash = iscc.instance_id(file.file)

    if not norm_title:
        iscc_code = "-".join((cid, did, iid))
    else:
        iscc_code = "-".join((mid, cid, did, iid))

    components = iscc_split(iscc_code)

    result = dict(
        iscc=iscc_code,
        tophash=tophash,
        gmt=gmt,
        bits=[code_to_bits(c) for c in components],
    )
    if norm_title:
        result["title"] = title
        result["title_trimmed"] = norm_title
    if norm_extra:
        result["extra"] = extra
        result["extra_trimmed"] = norm_extra

    file.file.close()
    return result


@app.post(
    "/generate/from_url",
    response_model=ISCC,
    tags=["generate"],
    summary="Generate ISCC from URL",
)
def from_url(url: HttpUrl):
    """Generate Full ISCC from URL."""
    result = iscc_from_url(url, guess=True)
    result["title"] = result.pop("norm_title")
    result["title_trimmed"] = result["title"]
    components = iscc_split(result["iscc"])
    result["bits"] = [code_to_bits(c) for c in components]
    return result


@app.post(
    "/generate/meta_id/",
    response_model=MetaID,
    response_model_exclude_unset=True,
    tags=["generate"],
    summary="Generate ISCC Meta-ID",
)
def meta_id(meta: Metadata):
    """Generate MetaID from 'title' and optional 'extra' metadata"""
    extra = meta.extra or ""
    mid, title_trimmed, extra_trimmed = iscc.meta_id(meta.title, extra)
    result = {
        "code": mid,
        "bits": code_to_bits(mid),
        "ident": code_to_int(mid),
        "title": meta.title,
        "title_trimmed": title_trimmed,
    }

    if extra:
        result["extra"] = extra
        result["extra_trimmed"] = extra_trimmed

    return result


@app.post(
    "/generate/content_id_text",
    response_model=ContentID,
    tags=["generate"],
    summary="Generate ISCC Content-ID-Text",
)
def content_id_text(text: Text):
    """Generate ContentID-Text from 'text'"""
    cid_t = iscc.content_id_text(text.text)
    return {
        "gmt": "text",
        "bits": code_to_bits(cid_t),
        "code": cid_t,
        "ident": code_to_int(cid_t),
    }


@app.post(
    "/generate/data_id",
    response_model=DataID,
    tags=["generate"],
    summary="Generate ISCC Data-ID",
)
def data_id(file: UploadFile = File(...)):
    """Generate Data-ID from raw binary data"""
    did = iscc.data_id(file.file)
    return {"code": did, "bits": code_to_bits(did), "ident": code_to_int(did)}


@app.post(
    "/generate/instance_id",
    response_model=InstanceID,
    tags=["generate"],
    summary="Generate ISCC Instance-ID",
)
def instance_id(file: UploadFile = File(...)):
    """Generate Instance-ID from raw binary data"""
    iid, tophash = iscc.instance_id(file.file)
    return {
        "code": iid,
        "bits": code_to_bits(iid),
        "ident": code_to_int(iid),
        "tophash": tophash,
    }


@app.post(
    "/generate/data_instance_id",
    tags=["generate"],
    summary="Generate ISCC Data-ID and Instance-ID",
)
def data_and_instance_id(file: UploadFile = File(...,)):
    """Generate Data-ID and Instance-ID from raw binary data"""

    did = iscc.data_id(file.file)
    file.file.seek(0)
    iid, tophash = iscc.instance_id(file.file)
    return {
        "data_id": {"code": did, "bits": code_to_bits(did), "ident": code_to_int(did),},
        "instance_id": {
            "code": iid,
            "bits": code_to_bits(iid),
            "ident": code_to_int(iid),
            "tophash": tophash,
        },
    }


@app.get(
    "/lookup",
    response_model=List[StreamItem],
    tags=["lookup"],
    summary="Lookup ISCC Codes",
)
def lookup(iscc: str):
    """Lookup an ISCC Code"""
    client = get_client()
    if client is None:
        raise HTTPException(
            HTTP_503_SERVICE_UNAVAILABLE, "ISCC lookup service not available"
        )
    try:
        iscc_verify(iscc)
    except ValueError as e:
        raise HTTPException(HTTP_400_BAD_REQUEST, str(e))

    components = iscc_split(iscc)
    results = []
    seen = set()
    for component in components:
        response = client.liststreamkeyitems(ISCC_STREAM, component, True, 100, 0, True)
        for result in response:
            txid = result.get("txid")
            if txid is None or txid in seen:
                continue
            results.append(result)
            seen.add(txid)
    result = stream_filter.search(results)
    cleaned = []
    for entry in result:
        keys = entry["keys"]
        # Better be conservative until we have a similarity based index.
        # So for now we only match if at least two components are identical.
        matches = set(keys).intersection(set(components))
        if not len(matches) >= 2:
            continue
        keys = add_placeholders(keys)

        entry["bits"] = [code_to_bits(c) for c in keys]
        while len(entry["bits"]) < 4:
            entry["bits"].append("0" * 64)
        cleaned.append(entry)
    return cleaned


def run_server():
    uvicorn.run("iscc_service.main:app", host="127.0.0.1", port=8000, reload=False)


if __name__ == "__main__":
    uvicorn.run("iscc_service.main:app", host="127.0.0.1", port=8000, reload=True)



# -*- coding: utf-8 -*-
from loguru import logger
import mcrpc

from iscc_service.config import NODE_IP, NODE_PORT, NODE_USER, NODE_PWD


client = None


def get_client():
    global client
    if client is not None:
        return client
    conn = mcrpc.RpcClient(NODE_IP, NODE_PORT, NODE_USER, NODE_PWD)
    try:
        info = conn.getblockchaininfo()
        logger.debug("Connected to {}".format(info))
        client = conn
        return conn
    except Exception as e:
        logger.warning("No connection to blockchain node: {}".format(e))
        return None


if __name__ == "__main__":
    get_client()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['iscc_service']

package_data = \
{'': ['*']}

install_requires = \
['bitstring>=3.1,<4.0',
 'fastapi>=0.54,<0.55',
 'iscc-cli>=0.9,<0.10',
 'jmespath>=0.9,<0.10',
 'loguru>=0.4,<0.5',
 'mcrpc>=2.0,<3.0',
 'python-multipart>=0.0.5,<0.0.6',
 'uvicorn>=0.11,<0.12']

entry_points = \
{'console_scripts': ['isccservice = iscc_service.main:run_server']}

setup_kwargs = {
    'name': 'iscc-service',
    'version': '0.2.1',
    'description': 'ISCC Web Service API',
    'long_description': "# iscc-service - ISCC Web Service API\n[![Version](https://img.shields.io/pypi/v/iscc-service.svg)](https://pypi.python.org/pypi/iscc-service/)\n[![Downloads](https://pepy.tech/badge/iscc-service)](https://pepy.tech/project/iscc-service)\n[![Pulls](https://shields.beevelop.com/docker/pulls/titusz/iscc-service.svg?style=flat-square)](https://hub.docker.com/r/titusz/iscc-service)\n\n> A REST OpenAPI Backend for creating [**ISCC codes**](https://iscc.codes) for digital media files.\n\n\n**Note**: This is work in progress. Use with care and at your own risk\n\nThe Webservice is build with [FastAPI](https://github.com/tiangolo/fastapi) and makes\nuse of the [ISCC reference implementation](<https://github.com/iscc/iscc-specs>) and\nthe [ISCC Command Line Tool](https://github.com/iscc/iscc-cli) and includes an\ninteractive API documentation:\n\n![Interactive ISCC Api Docs](screenshot.jpg)\n\n\nThe Docker image is published at https://hub.docker.com/r/titusz/iscc-service\n\n\n## Setup for development\n\nIf you are using [poetry](https://python-poetry.org/):\n\n- After checkout cd into code directory and run 'poetry install' to install dependencies.\n- Launch dev server with: 'uvicorn iscc_service.main:app --reload'\n- See API docs at: http://127.0.0.1:8000\n\nFor the 'lookup' endpoint to work you must provide env variables for node connection.\nSee [config.py](iscc_service/config.py)\n\n## Install via pip\n\n```bash\n$ pip3 install iscc-service\n```\n\nRun webservice via uvicorn\n\n```bash\n$ isccservice\nINFO:     Started server process [18800]\nINFO:     Waiting for application startup.\nINFO:     Application startup complete.\nINFO:     Uvicorn running on http://127.0.0.1:8000 (Press CTRL+C to quit)\n```\n\n## Publishing on Docker HUB:\n\n```bash\ndocker login\ndocker build -t iscc-service -f Dockerfile .\ndocker run --rm -p 8080:8080 -it iscc-service\ndocker tag iscc-service username/iscc-service:0.1.9\ndocker push username/iscc-service:0.1.9\n```\n\n## Change Log\n\n### [0.2.1] - 2020-05-13\n- Update to iscc-cli 0.9.8\n- More conservative lookup matching\n\n### [0.2.0] - 2020-05-01\n- Update to support flac and opus audio files\n\n### [0.1.9] - 2020-04-27\n\n- Support updated Content-ID Audio\n- Support incomplete ISCC codes\n- Updated dependencies\n\n### [0.1.8] - 2020-03-02\n\n- Add support for mobi files\n- Initial pypi release\n\n\n## License\n\nMIT © 2019-2020 Titusz Pan\n",
    'author': 'Titusz Pan',
    'author_email': 'tp@py7.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://iscc.codes/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
import iscc
import jmespath
from bitstring import BitArray


def code_to_bits(code: str) -> str:
    """Convert ISCC Code to bitstring"""
    data = iscc.decode(code)
    ba = BitArray(data[1:])
    return ba.bin


def code_to_int(code: str) -> int:
    """Convert ISCC Code to integer"""
    data = iscc.decode(code)
    ba = BitArray(data[1:])
    return ba.uint


stream_filter = jmespath.compile(
    "[].{txid: txid, vout: vout, keys: keys, title: data.json.title, tophash:data.json.tophash, extra:data.json.extra, time: time, content_url: data.json.meta[0].data.encoding[0].contentUrl}"
)


def add_placeholders(components):
    """Add placeholders for missing DATA/INSTANCE components"""
    headers = [s[:2] for s in components]
    for prefix in ("CD", "CR"):
        if prefix not in headers:
            components.append(prefix + ("C" * 11))
    return components


if __name__ == "__main__":
    print(code_to_bits("CCDFPFc87MhdT"))
    print(code_to_int("CCDFPFc87MhdT"))
